from __future__ import absolute_import, division, print_function
from maya import cmds, mel
import maya.api.OpenMaya as om
from vray import material_converter, vray_maya_utils

''' LINEAR WORKFLOW '''

'''
Map of nodes and their respective input attributes used in the linearWorkflow migration.
For each node there's a respective map containing the color attribute that should be processed.
The other part of the map is the amount/multiplier tied to the color if present, otherwise - empty string.
'''
_migrateNodeAttrList = {
	'VRayCarPaintMtl' : {
		'color' : ''
	},
	'VRayMtl' : {
		'color' : 'diffuseColorAmount',
		'reflectionColor' : 'reflectionColorAmount',
		'refractionColor' : 'refractionColorAmount',
		'translucencyColor' : '',
	},
	'VRayToonMtl' : {
		'diffuseColor' : 'diffuseColorAmount',
		'reflectionColor' : 'reflectionColorAmount',
		'refractionColor' : 'refractionColorAmount',
		'translucencyColor' : '',
	},
	'VRaySkinMtl' : {
		'shallowColor' : '',
		'mediumColor' : '',
		'deepColor' : '',
		'diffuseColor' : '',
		'primaryReflectonColor' : '',
		'secondaryReflectonColor' : '',
	},
	'VRayFastSSS2' : {
		'diffuseTex' : '',
		'overallTex' : '',
		'reflection' : '',
		'scatterRadiusColor' : '',
		'subsurfaceColor' : ''
	}
}

def isclose(a, b):
    '''
    Are two floating point numbers within 1e-6 of each other
    '''
    return -1e-6 < a - b < 1e-6

class LinearWorkflowMigrator():
    ''' Manages the linear workflow migration process. '''
    
    # map of nodes and their respective input attributes used in the linear workflow migration
    nodeAttrMap = {}
    
    def __init__(self, nodeAttrlist):
        '''
        :param nodeAttrlist: list with migration nodes and their attributes that should be processed
        '''
        self.nodeAttrMap = nodeAttrlist
    
    def applyDoubleGamma(self, nodeName, attrName):
        '''
        Check if double gamma should be applied for the given attribute.
        :param nodeName: Name of the parent node
        :param attrName: Color attribute name
        :rtype: Boolean
        '''
        nodeType = cmds.nodeType(nodeName)
        allowedDoubleGamma = nodeType=='VRayFastSSS2' and attrName=='diffuseTex'
        # the local linearWorkflow option is taken into account only if the global is enabled (which is the only case considered)
        return allowedDoubleGamma and cmds.getAttr(nodeName+'.linearWorkflow')
    
    def getNodeTypes(self):
        ''' Return the node types from the map. '''
        return self.nodeAttrMap.keys()
    
    def getAmountAttr(self, node, colorAttr):
        ''' Return the corresponding amount attribute name if present, otherwise - empty string. '''
        nodeType = cmds.nodeType(node)
        notFound = nodeType not in self.nodeAttrMap or colorAttr not in self.nodeAttrMap[nodeType]
        if notFound:
            om.MGlobal.displayWarning('Could not find attribute "{}" from node "{}" in the migration map'.format(colorAttr, node))
            return ''
        return self.nodeAttrMap[nodeType][colorAttr]
     
    def processPlug(self, attrPlug, nodeName, attrName):
        '''
        Check if the given plug is valid to be processed. If it satisfies the needed conditions,
        a respective gamma node will be further created for the plug with an optional node
        to handle the case of a color attribute with a corresponding amount.
        :param attrPlug: MPlug for a color attribute which should be checked.
        :param nodeName: Name of the node.
        :param attrName: Color attribute name as described in the map. It might be different
          from the plug's name for aliases. Then the plug points to the original attribute,
          while we store the alias name in the map.
        :rtype: Boolean
        '''
        if not isinstance(attrPlug, om.MPlug) or attrPlug.isNull:
            raise TypeError('Expected MPlug argument')
            return False
        # consider only color attributes
        if not cmds.attributeQuery(attrName, node=nodeName, at=True)=='float3':
            return False
        isSimpleColor = not attrPlug.isDestination
        if isSimpleColor:
            r = attrPlug.child(0).asFloat()
            g = attrPlug.child(1).asFloat()
            b = attrPlug.child(2).asFloat()
            isBlack = isclose(r,0) and isclose(g,0) and isclose(b,0)
            isWhite = isclose(r,1) and isclose(g,1) and isclose(b,1)
            # it makes no sense to apply correction on pure white and black colors (no effect in this case)
            # and fill the scene with unnecessary nodes, so skip those
            if (isBlack or isWhite) and not self.needsMultiplyDivide(nodeName, attrName):
                return False
        return True
    
    def needsMultiplyDivide(self, node, colorAttr):
        '''
        Check if a "multiplyDivide" node should be created. This will be done
        for a color attribute which has an amount that is either connected or !=1.
        :param node: Name of the node
        :param colorAttr: Name of the color attribute
        :rtype: Boolean
        '''
        amountName = self.getAmountAttr(node, colorAttr)
        if not amountName:
            return False
        depNode = om.MFnDependencyNode(material_converter.find_by_name(node))
        amountPlug = depNode.findPlug(amountName, True)
        isConnected = not amountPlug.isNull and amountPlug.isDestination
        return isConnected or not isclose(cmds.getAttr(node+'.'+amountName), 1)
    
    def transferColorPlug(self, inputPlug, nodeName, inputAttrName, dstInputAttr, dstOutputAttr):
        '''
        Copy the input plug to a destination node. This includes value copying for simple colors or
        disconnecting a texture connection and connecting it to the destination node.
        :param inputPlug: MPlug for the original color attribute
        :param nodeName: Name of the node
        :param inputAttrName: Color's partial attribute name as described in the map. It might be different
          from the plug's name for aliases. Then the plug points to the original attribute,
          while we store the alias name in the map.
        :param dstInputAttr: Name of the destination input attribute where the value will be copied
        :param dstOutputAttr: Name of the destination output attribute which will be connected to the original color
        '''
        if not isinstance(inputPlug, om.MPlug) or inputPlug.isNull:
            raise TypeError('Expected MPlug argument')
            return
        attrType = cmds.attributeQuery(inputAttrName, node=nodeName, at=True)
        if attrType != 'float3':
            return
        if inputPlug.isDestination:
            connectedTexPlug = inputPlug.source()
            if not connectedTexPlug.isNull:
                cmds.disconnectAttr(connectedTexPlug.name(), inputPlug.name())
                cmds.connectAttr(connectedTexPlug.name(), dstInputAttr, force=True)
        else:
            r = inputPlug.child(0).asFloat()
            g = inputPlug.child(1).asFloat()
            b = inputPlug.child(2).asFloat()
            cmds.setAttr(dstInputAttr, r, g, b, type='float3')
        cmds.connectAttr(dstOutputAttr, inputPlug.name(), force=True)
        
    def transferFloatPlug(self, inputPlug, dstInputAttr):
        '''
        Copy the input plug to a destination node. This includes float value copying (if not connected) or
        disconnecting a texture connection and connecting it to the destination node.
        The latter is supposed to be "multiplyDivide" node.
        :param inputPlug: MPlug for the original float attribute
        :param dstInputAttr: Name of the destination input attribute where the value will be copied
        '''
        if not isinstance(inputPlug, om.MPlug) or inputPlug.isNull:
            raise TypeError('Expected MPlug argument')
            return
        nodeName = mel.eval('vrayGetPlugNodeName("'+inputPlug.name()+'")')
        attrType = cmds.attributeQuery(inputPlug.partialName(useLongNames=True), node=nodeName, at=True)
        if attrType != 'float':
            return
        if inputPlug.isDestination:
            connectedTexPlug = inputPlug.source()
            connectedTexPlugName = connectedTexPlug.name()
            nodeName = mel.eval('vrayGetPlugNodeName("'+connectedTexPlug.name()+'")')
            isColor = cmds.attributeQuery(connectedTexPlug.partialName(useLongNames=True), node=nodeName, at=True)=='float3'
            cmds.disconnectAttr(connectedTexPlugName, inputPlug.name())
            if isColor:
                # plug directly
                cmds.connectAttr(connectedTexPlugName, dstInputAttr, force=True)
            else:
                # plug the float connection in each of the 3 slots
                cmds.connectAttr(connectedTexPlugName, dstInputAttr+'X', force=True)
                cmds.connectAttr(connectedTexPlugName, dstInputAttr+'Y', force=True)
                cmds.connectAttr(connectedTexPlugName, dstInputAttr+'Z', force=True)
        else:
            value = cmds.getAttr(inputPlug.name())
            cmds.setAttr(dstInputAttr, value, value, value, type='float3')
        cmds.setAttr(inputPlug.name(), 1)
    
    def doMigration(self):
        '''
        Linear workflow migration script.
        :return String containing all messages providing information about the result of the script.
        '''
        outMsg = ''
        mel.eval('vrayCreateVRaySettingsNode')
        if not canMigrateLinearWorkflow():
            return outMsg
        
        gammaValue = cmds.getAttr('vraySettings.cmap_gamma')
        # if the gamma value is 1, then only disable linearWorkflow option
        if isclose(gammaValue, 1):
            cmds.setAttr('vraySettings.cmap_linearworkflow', 0)
            return outMsg
        
        gammaValue = 1/gammaValue
        allNodeTypes = self.getNodeTypes()
        migrateNodes = cmds.ls(type=allNodeTypes)
        
        # iterate all scene nodes that should be migrated
        for node in migrateNodes:
            nodeType = cmds.nodeType(node)
            depNode = om.MFnDependencyNode(material_converter.find_by_name(node))
            for colorAttr in self.nodeAttrMap[nodeType]:
                colorPlug = om.MPlug()
                if depNode.hasAttribute(colorAttr):
                    colorPlug = depNode.findPlug(colorAttr, True)
                else:
                    # handle alias attributes
                    mObj = depNode.findAlias(colorAttr)
                    if not mObj.isNull():
                        colorPlug = depNode.findPlug(mObj, True)
                
                if self.processPlug(colorPlug, node, colorAttr):
                    gammaNodeName = cmds.shadingNode('gammaCorrect', asUtility=True)
                    cmds.setAttr(gammaNodeName+'.gamma', gammaValue, gammaValue, gammaValue, type='float3')
                    gammaInputAttr = gammaNodeName+'.value'
                    gammaOutputAttr = gammaNodeName+'.outValue'
                    if self.applyDoubleGamma(node, colorAttr):
                        gammaNode2Name = cmds.shadingNode('gammaCorrect', asUtility=True)
                        cmds.setAttr(gammaNode2Name+'.gamma', gammaValue, gammaValue, gammaValue, type='float3')
                        cmds.connectAttr(gammaOutputAttr, gammaNode2Name+'.value', force=True)
                        gammaOutputAttr = gammaNode2Name+'.outValue'
                    
                    self.transferColorPlug(colorPlug, node, colorAttr, gammaInputAttr, gammaOutputAttr)
                    
                    if self.needsMultiplyDivide(node, colorAttr):
                        amountAttr = self.getAmountAttr(node, colorAttr)
                        amountPlug = depNode.findPlug(amountAttr, True)
                        multDivNode = cmds.shadingNode('multiplyDivide', asUtility=True)
                        gammaDepNode = om.MFnDependencyNode(material_converter.find_by_name(gammaNodeName))
                        gammaInputPlug = gammaDepNode.findPlug('value', True)
                        
                        self.transferColorPlug(gammaInputPlug, gammaNodeName, 'value', multDivNode+'.input1', multDivNode+'.output')
                        self.transferFloatPlug(amountPlug, multDivNode+'.input2')
        
        cmds.setAttr('vraySettings.cmap_linearworkflow', 0)
        msg = 'Legacy Linear Workflow is replaced by gammaCorrect nodes.\n'
        outMsg += msg
        return outMsg

def canMigrateLinearWorkflow():
    ''' Preliminary check which is True if the linear workflow migration should be run. '''
    return cmds.objExists('vraySettings') and cmds.getAttr('vraySettings.cmap_linearworkflow')
    
def migrateLinearWorkflow():
    '''
    Run Linear workflow migration script.
    :return String containing all messages providing information about the result of the script.
    '''
    
    lwfMigrator = LinearWorkflowMigrator(_migrateNodeAttrList)
    return lwfMigrator.doMigration()

migrateLinearWorkflow()